// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

/* -*-antlr-*- */
header {

    package de.uka.ilkd.key.parser;

    import java.io.*;
    import antlr.*;
    import de.uka.ilkd.key.logic.*;
    import de.uka.ilkd.key.logic.sort.*;    
    import de.uka.ilkd.key.logic.op.*;
    import de.uka.ilkd.key.rule.metaconstruct.*;
    import de.uka.ilkd.key.rule.*;
    import de.uka.ilkd.key.proof.init.Includes;
    import java.util.List;
    import java.util.LinkedList;
    import java.util.HashMap;
    import java.util.HashSet;
    import java.util.Iterator;
    import de.uka.ilkd.key.java.visitor.ProgramVariableCollector;
    import de.uka.ilkd.key.java.Recoder2KeY;
    import de.uka.ilkd.key.java.JavaReader;
    import de.uka.ilkd.key.java.JavaInfo;
    import de.uka.ilkd.key.java.Services;
    import de.uka.ilkd.key.java.recoderext.*;
    import de.uka.ilkd.key.java.abstraction.KeYJavaType;
    import de.uka.ilkd.key.proof.UseInsteadMap;
    import de.uka.ilkd.key.proof.RuleSource;
    import de.uka.ilkd.key.gui.AbbrevMap;
    
}

/** ONLY FOR TEST CASES.
 * A parser for first-order terms and formulae.  Variables in 
 * quantifiers are expected to be declared globally in the variable
 * namespace.  This parser is used for test cases, where you want to know
 * in advance which objects will represent bound variables.
 */  

class GlobalDeclarationTermParser extends TermParser;
options {
	importVocab=KeYLexer;
	k = 1;
	defaultErrorHandler=true;
}


{
    static final Sort[] AN_ARRAY_OF_SORTS = new Sort[0];
    static final Term[] AN_ARRAY_OF_TERMS = new Term[0];

    protected NamespaceSet nss;

    protected Choice defaultChoice = null; 
    protected HashMap category2Default = new HashMap();
    protected boolean parse_includes = false;
    protected Includes includes = new Includes();
    protected boolean onlyWith=false;
    protected SetOfChoice activatedChoices = SetAsListOfChoice.EMPTY_SET;
    protected HashSet usedChoiceCategories = new HashSet();

    protected AbbrevMap scm = new AbbrevMap();
	private Prioritization.StaticPrioritization heur_prioritization
		= new Prioritization.StaticPrioritization ();

    // these variables are set if a file is read in step by
    // step. This used when reading in LDTs because of cyclic
    // dependencies.
    protected boolean skip_functions;
    protected boolean skip_predicates;
    protected boolean skip_sorts;
    protected boolean skip_schemavariables;
    protected boolean skip_heuristics;
    protected UseInsteadMap useInstead=new UseInsteadMap();

    private boolean schemaMode;

    protected JavaReader javaReader;
    protected TermFactory tf;
    private Services services = new Services();

    /** collects errors and warnings that occured during parsing */
    protected ErrorHandler errHandler = new ErrorHandler();

    public boolean errors = false;



    public GlobalDeclarationTermParser(TokenStream lexer,
				       TermFactory tf,
		  		       JavaReader jr,
				       NamespaceSet nss,
                       Choice defaultChoice) {
	this(lexer,2);
	this.tf = tf;
	this.javaReader = jr;
	this.nss     = nss;
	switchToNormalMode();
    this.defaultChoice = defaultChoice;
    }

    public GlobalDeclarationTermParser(TokenStream lexer,
				       TermFactory tf,
		  		       Services services,
				       NamespaceSet nss,
                       Choice defaultChoice) {
	this(lexer, tf, 
		new Recoder2KeY(new KeYCrossReferenceServiceConfiguration(), 
				services.getJavaInfo().rec2key(), new NamespaceSet(), 
				services.getTypeConverter()),
   	        nss, defaultChoice);	
    }

    public GlobalDeclarationTermParser(TokenStream lexer,
				       Services services,
				       NamespaceSet nss) {
	this(lexer, TermFactory.DEFAULT,
	     new Recoder2KeY(new KeYCrossReferenceServiceConfiguration(), 
		 services.getJavaInfo().rec2key(), new NamespaceSet(), services.getTypeConverter()),
	        nss,new Choice(new Name("Default")));
    }
    
    private void switchToNormalMode() {
        schemaMode   = false;
    }

    private boolean inSchemaMode() {
	return schemaMode;
    }

    private void switchToSchemaMode() {
        schemaMode   = true;
    }

    public NamespaceSet namespaces(){
    return nss;
    }

    public Namespace sorts() {
	return nss.sorts();
    }

    public Namespace variables() {
	return nss.variables();
    }

    public Namespace functions() {
	return nss.functions();
    }

    public Namespace programVariables() {
        return nss.programVariables();
    }

    public Namespace heuristics() {
	return nss.heuristics();
    }

    public Namespace choices() {
	return nss.choices();
    }

    public Named lookup(Name n) {
	final Namespace[] lookups = {
		functions(), variables(), 
		programVariables()
	};

	for (int i = 0; i<lookups.length; i++) {
	    if (lookups[i].lookup(n) != null) {
		return lookups[i].lookup(n);
            }
	}
	return null;	
    }

    public Term toZNotation(String s){
        Term result = tf.createFunctionTerm
	    ( (Function) functions().lookup(new Name(s.substring(0,1))),
	      tf.createFunctionTerm((Function) functions().lookup(new Name("#"))));

        for(int i=1;i<s.length();i++){
	    result = tf.createFunctionTerm
		((Function)functions().lookup
		  (new Name(s.substring(i,i+1))), result);
        }
        return tf.createFunctionTerm
	    ((Function) functions().lookup(new Name("Z")), result);	
    }


    protected KeYJavaType getTypeByClassName(String s){
        KeYJavaType kjt = null;
        try {
            kjt = getJavaInfo().getTypeByClassName(s);
        } catch(RuntimeException e){
            return null;
        } 
        return kjt;
    }

    protected boolean isNumber(String s){
	      boolean num=true;
          if(s==null){
            num=false;
          }else{
	        for(int i=s.length()-1;i>=0;i--){
                if(s.charAt(i)<'0'||s.charAt(i)>'9')
			    num=false;
	        }
         }
	     return num;
    }


    protected int getLine() {
	int line = -1;
	try {
	    line = LT(0).getLine();
	} 
	catch (TokenStreamException e) {
	    System.err.println("No further token in stream");
	}
	return line;
    }   

    protected int getColumn() {
	int col = -1;
	try {
	    col = LT(0).getColumn();
	} 
	catch (TokenStreamException e) {
	    System.err.println("No further token in stream");
	}
	return col;
    }   

    // These three methods are required because the unused productions
    // from terms.g are inherited.  They should not be called.
    protected LogicVariable bindVar(Token id,Sort s) {
	throw new RuntimeException("bindVar was called in " +
				   "GlobalDeclarationTermParser.");
    }
	
    protected ProgramVariable progVar(Token id,Sort s) {
	variables().lookup(new Name(id.getText()));
	ProgramVariable v = (ProgramVariable)(ProgramVariable)
	    variables().lookup(new Name(id.getText()));
	if (v == null) {
	    v = new ProgramVariable(new ProgramElementName(id.getText()), s);
	    nss.setVariables(variables().extended(v));
	}
	return v;
    }

    protected void unbindVars() {
	throw new RuntimeException("unbindVars was called in " +
				   "GlobalDeclarationTermParser.");
    }


    protected static HashSet progVars(JavaBlock jb) {
	ProgramVariableCollector pvc
	    = new ProgramVariableCollector(jb.program());
	pvc.start();
	return pvc.result();
    }


    protected void bindProgVars(HashSet progVars) {
	nss.setVariables(new Namespace(variables()));
	Iterator it=progVars.iterator();
	while (it.hasNext()) {
	    variables().add((ProgramVariable)it.next());
	}
    }

    protected void unbindProgVars() {
	nss.setVariables(variables().parent());
    }


    protected Term termForParsedVariable(ParsableVariable v) 
	throws antlr.SemanticException 
	{
	    if ( v instanceof LogicVariable ) {
		return tf.createVariableTerm((LogicVariable)v);
	    } else if ( v instanceof ProgramVariable ) {
		return tf.createVariableTerm((ProgramVariable)v);
	    } else {
		throw new antlr.SemanticException(
						  ""+ v +" is not a logic variable");
	    }
	}

    protected void addInclude(String filename, boolean relativePath, boolean ldt){
        InputStream input=null;
        RuleSource source=null;
        if(relativePath){
            int end = getFilename().lastIndexOf(File.separator);
            int start = 0;
            filename = filename.replace('/', File.separatorChar);
            filename = filename.replace('\\', File.separatorChar);
            if(getFilename().startsWith("file:")){
                start = 5;
            }
            File path=new File(getFilename().substring(start,end+1)+filename);
            try{ 
                source = RuleSource.initRuleFile(path.toURL()); 
            }catch(java.net.MalformedURLException e){
                System.err.println("Exception due to malformed URL of file "+
                   filename+"\n " +e);
            }
            filename = filename.substring(
                filename.lastIndexOf(File.separator)+1,
                filename.lastIndexOf(".key"));
        }else{
            source = RuleSource.initRuleFile(filename+".key"); 
        }
        if(ldt){
            includes.putLDT(filename, source);
        }else{
            includes.put(filename, source);
        }
    }  

    protected JavaBlock getJavaBlock(Token t) {
	return javaReader.readBlockWithEmptyContext(t.getText());
    }

    /** returns the java information object */
    public JavaInfo getJavaInfo() {
	return null;
//	return javaReader.createJavaInfo(new Services());
    }

	public Services getServices() {
	    return services;
    	}
}


top {Term t;} : t=formula;

bindingterm returns [Term a = null]
	:
		a = global_bindingterm
	;

/** here is the special behaviour, that requires a bound variable
 * to be previously declared in <code>variables()</code>. */

global_bindingterm returns [Term a = null]
{
	Operator op = null;
	ParsableVariable v0, v1;
	LogicVariable v = null;
	ProgramVariable pv = null;
	Term a1,a2;
	boolean waryAll = false;
	boolean waryEx = false;
}
	:
		( 
			"all" { op = Op.ALL; }
		|   "ex"  { op = Op.EX;  }
		)
		
		v0=varId DOT // Don't need the sort here
		{v = (LogicVariable) v0;} a1=term60 
		{ 
			a = tf.createQuantifierTerm((Quantifier)op, v,a1);
		}
	|
		LBRACE v1=varId
		(
		a1=term // Don't need the sort here
		( "waryAll" { waryAll = true; } |
		  "waryEx"  { waryEx  = true; } )?
		RBRACE
		{v = (LogicVariable) v1;} a2=term60
		{ 
		    if ( waryAll )
		        a = tf.createSubstitutionTerm ( Op.WARY_SUBST_ALL,
		                                        v, a1, a2 );
		    else if ( waryEx )
		        a = tf.createSubstitutionTerm ( Op.WARY_SUBST_EX,
		                                        v, a1, a2 );
		    else
		        a = tf.createSubstitutionTerm ( Op.SUBST,
		                                        v, a1, a2 );
		}
		|
		// simple updates, just for tests
		ASSIGN a1=term RBRACE
		{pv = (ProgramVariable) v1;} a2=term60
		{ 
		    a = tf.createUpdateTerm ( new Term[] { tf.createVariableTerm ( pv ),
		                                           a1, a2 } );
		}
		)
	;
