// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.proof.mgt;

import de.uka.ilkd.key.proof.*;
import de.uka.ilkd.key.rule.Taclet;
import de.uka.ilkd.key.rule.NoPosTacletApp;
import java.util.LinkedList;
import java.util.List;
import de.uka.ilkd.key.proof.init.ProofOblInput;

public class LemmaSpec implements Contract {
    
    private NoPosTacletApp lemma;
    private ProofEnvironment env;
    private List proofs = new LinkedList();


    public LemmaSpec(NoPosTacletApp lemma) {
	this.lemma = lemma;
    }


    public void setProofEnv(ProofEnvironment env) {
	if (this.env==null) {
	    this.env=env;
	} else if (this.env!=env) {
	    throw new IllegalStateException("LemmaSpec may belong to "
					    +"only one environment.");
	}
    }

    public ProofEnvironment getProofEnv() {
	return env;
    }

    public Taclet getLemmaAndRegister() {
	return lemma.taclet();
    }

    public List getProofs() {
	return proofs;
    }

    public void addProofList(ProofList pl) {
	proofs.add(pl);
    }

    public void removeProofList(ProofList pl) {
	proofs.remove(pl);
    }


    public Object getObjectOfContract() {
	return lemma;
    }

    public ProofOblInput getProofOblInput() {
	return null; // not implemented yet
    }

    public String toString() {
	return "Lemma specification: "+lemma;
    }

    public boolean equals(Object cmp) {
	if (!(cmp instanceof LemmaSpec)) return false;
	LemmaSpec lc = (LemmaSpec)cmp;
	return (lemma.equals(lc.lemma));
    }

    public int hashCode() {
	return lemma.hashCode();
    }

}
